"""Tests for dataset.select_rows(udf_col)."""

from typing import ClassVar, Iterable, Iterator, Optional, cast

import numpy as np
import pytest
from pytest import approx
from typing_extensions import override

from ..concepts.concept import ExampleIn
from ..concepts.db_concept import ConceptUpdate, DiskConceptDB
from ..schema import (
  ROWID,
  SPAN_KEY,
  Field,
  Item,
  RichData,
  SignalInputType,
  SpanVector,
  field,
  lilac_embedding,
  span,
)
from ..signal import (
  TextEmbeddingSignal,
  TextSignal,
  VectorSignal,
  clear_signal_registry,
  register_signal,
)
from ..signals.concept_scorer import ConceptSignal
from .dataset import BinaryFilterTuple, Column, SortOrder
from .dataset_test_utils import TestDataMaker, enriched_item

EMBEDDINGS: list[tuple[str, list[float]]] = [
  ('hello.', [1.0, 0.0, 0.0]),
  ('hello2.', [1.0, 1.0, 0.0]),
  ('hello world.', [1.0, 1.0, 1.0]),
  ('hello world2.', [2.0, 1.0, 1.0]),
]

STR_EMBEDDINGS: dict[str, list[float]] = {text: embedding for text, embedding in EMBEDDINGS}


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      if example == '':
        yield None
        continue
      yield [lilac_embedding(0, len(example), np.array(STR_EMBEDDINGS[cast(str, example)]))]


class LengthSignal(TextSignal):
  name: ClassVar[str] = 'length_signal'

  _call_count: int = 0

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      self._call_count += 1
      yield len(text_content)


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  @override
  def fields(self) -> Field:
    return field(fields={'len': 'int32', 'flen': 'float32'})

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    return ({'len': len(text_content), 'flen': float(len(text_content))} for text_content in data)


class TestEmbeddingSumSignal(VectorSignal):
  """Sums the embeddings to return a single floating point value."""

  name: ClassVar[str] = 'test_embedding_sum'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  @override
  def fields(self) -> Field:
    return field('float32')

  @override
  def vector_compute(
    self, all_vector_spans: Iterable[list[SpanVector]]
  ) -> Iterator[Optional[Item]]:
    # The signal just sums the values of the embedding.
    for vector_spans in all_vector_spans:
      yield vector_spans[0]['vector'].sum()


class ComputedKeySignal(TextSignal):
  name: ClassVar[str] = 'computed_key'

  @override
  def fields(self) -> Field:
    return field('int64')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text in data:
      yield 1

  def key(self, is_computed_signal: Optional[bool] = False) -> str:
    return f'key_{is_computed_signal}'


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_signal(LengthSignal)
  register_signal(TestSplitter)
  register_signal(TestEmbedding)
  register_signal(TestSignal)
  register_signal(TestEmbeddingSumSignal)
  register_signal(ComputedKeySignal)

  # Unit test runs.
  yield
  # Teardown.
  clear_signal_registry()


def test_udf(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])

  signal_col = Column('text', signal_udf=TestSignal())
  result = dataset.select_rows(['text', signal_col])

  assert list(result) == [
    {'text': 'hello', 'text.test_signal': {'len': 5, 'flen': 5.0}},
    {'text': 'everybody', 'text.test_signal': {'len': 9, 'flen': 9.0}},
  ]


def test_udf_with_filters(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])

  signal_col = Column('text', signal_udf=TestSignal())
  # Filter by source feature.
  filters: list[BinaryFilterTuple] = [('text', 'equals', 'everybody')]
  result = dataset.select_rows(['text', signal_col], filters=filters)
  assert list(result) == [{'text': 'everybody', 'text.test_signal': {'len': 9, 'flen': 9.0}}]


def test_udf_with_rowid_filter(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])

  # Filter by a specific rowid.
  filters: list[BinaryFilterTuple] = [(ROWID, 'equals', '1')]
  udf_col = Column('text', signal_udf=LengthSignal())
  result = dataset.select_rows([ROWID, 'text', udf_col], filters=filters)
  assert list(result) == [{ROWID: '1', 'text': 'hello', 'text.length_signal': 5}]
  assert cast(LengthSignal, udf_col.signal_udf)._call_count == 1

  filters = [(ROWID, 'equals', '2')]
  result = dataset.select_rows([ROWID, 'text', udf_col], filters=filters)
  assert list(result) == [{ROWID: '2', 'text': 'everybody', 'text.length_signal': 9}]
  assert cast(LengthSignal, udf_col.signal_udf)._call_count == 1 + 1

  # No filters.
  result = dataset.select_rows([ROWID, 'text', udf_col])
  assert list(result) == [
    {ROWID: '1', 'text': 'hello', 'text.length_signal': 5},
    {ROWID: '2', 'text': 'everybody', 'text.length_signal': 9},
  ]
  assert cast(LengthSignal, udf_col.signal_udf)._call_count == 2 + 2


def test_udf_with_rowid_filter_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': ['hello', 'hi']}, {'text': ['everybody', 'bye', 'test']}])

  # Filter by a specific rowid.
  filters: list[BinaryFilterTuple] = [(ROWID, 'equals', '1')]
  udf_col = Column(('text', '*'), signal_udf=LengthSignal())
  result = dataset.select_rows([ROWID, 'text', udf_col], filters=filters)
  assert list(result) == [{ROWID: '1', 'text': ['hello', 'hi'], 'text.length_signal': [5, 2]}]
  assert cast(LengthSignal, udf_col.signal_udf)._call_count == 2

  # Filter by a specific rowid.
  filters = [(ROWID, 'equals', '2')]
  result = dataset.select_rows([ROWID, 'text', udf_col], filters=filters)
  assert list(result) == [
    {ROWID: '2', 'text': ['everybody', 'bye', 'test'], 'text.length_signal': [9, 3, 4]}
  ]
  assert cast(LengthSignal, udf_col.signal_udf)._call_count == 2 + 3


def test_udf_deeply_nested(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': [['hello'], ['hi', 'bye']]}, {'text': [['everybody', 'bye'], ['test']]}]
  )

  udf_col = Column(('text', '*', '*'), signal_udf=LengthSignal())
  result = dataset.select_rows([udf_col])
  assert list(result) == [
    {'text.length_signal': [[5], [2, 3]]},
    {'text.length_signal': [[9, 3], [4]]},
  ]
  assert cast(LengthSignal, udf_col.signal_udf)._call_count == 6


def test_udf_with_embedding(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  dataset.compute_signal(TestEmbedding(), 'text')

  signal_col = Column('text', signal_udf=TestEmbeddingSumSignal(embedding='test_embedding'))
  result = dataset.select_rows(['text', signal_col])

  expected_result: list[Item] = [
    {'text': 'hello.', 'text.test_embedding_sum(embedding=test_embedding)': 1.0},
    {'text': 'hello2.', 'text.test_embedding_sum(embedding=test_embedding)': 2.0},
  ]
  assert list(result) == expected_result

  # Select rows with alias.
  signal_col = Column(
    'text', signal_udf=TestEmbeddingSumSignal(embedding='test_embedding'), alias='emb_sum'
  )
  result = dataset.select_rows(['text', signal_col])
  expected_result = [{'text': 'hello.', 'emb_sum': 1.0}, {'text': 'hello2.', 'emb_sum': 2.0}]
  assert list(result) == expected_result


def test_udf_with_nested_embedding(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': ['hello.', 'hello world.']}, {'text': ['hello world2.', 'hello2.']}]
  )

  dataset.compute_signal(TestEmbedding(), ('text', '*'))

  signal_col = Column(('text', '*'), signal_udf=TestEmbeddingSumSignal(embedding='test_embedding'))
  result = dataset.select_rows([('text', '*'), signal_col])
  expected_result = [
    {
      'text': ['hello.', 'hello world.'],
      'text.test_embedding_sum(embedding=test_embedding)': [1.0, 3.0],
    },
    {
      'text': ['hello world2.', 'hello2.'],
      'text.test_embedding_sum(embedding=test_embedding)': [4.0, 2.0],
    },
  ]
  assert list(result) == expected_result


def test_udf_throws_without_precomputing(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  # Embedding is not precomputed, yet we ask for the embedding.

  signal_col = Column('text', signal_udf=TestEmbeddingSumSignal(embedding='test_embedding'))

  with pytest.raises(ValueError, match="No embedding found for path \\('text',\\)"):
    dataset.select_rows(['text', signal_col])


class TestSplitter(TextSignal):
  """Split documents into sentence by splitting on period."""

  name: ClassVar[str] = 'test_splitter'

  @override
  def fields(self) -> Field:
    return field(fields=['string_span'])

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    for text in data:
      if not isinstance(text, str):
        raise ValueError(f'Expected text to be a string, got {type(text)} instead.')
      result: list[Item] = []
      for sentence in text.split('.'):
        start = text.index(sentence)
        end = start + len(sentence)
        result.append(span(start, end))
      yield result


def test_udf_after_precomputed_split(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': 'sentence 1. sentence 2 is longer'}, {'text': 'sentence 1 is longer. sent2 is short'}]
  )
  dataset.compute_signal(TestSplitter(), 'text')
  udf = Column('text', signal_udf=LengthSignal())
  result = dataset.select_rows(['*', udf], combine_columns=True)
  assert list(result) == [
    {
      'text': enriched_item(
        'sentence 1. sentence 2 is longer',
        {'length_signal': 32, 'test_splitter': [span(0, 10), span(11, 32)]},
      )
    },
    {
      'text': enriched_item(
        'sentence 1 is longer. sent2 is short',
        {'length_signal': 36, 'test_splitter': [span(0, 20), span(21, 36)]},
      )
    },
  ]


def test_is_computed_signal_key(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  signal_col = Column('text', signal_udf=ComputedKeySignal())
  result = dataset.select_rows(['text', signal_col])
  assert list(result) == [
    {'text': 'hello.', 'text.key_False': 1},
    {'text': 'hello2.', 'text.key_False': 1},
  ]


def test_sparse_embedding_index_with_asc_sort(make_test_data: TestDataMaker) -> None:
  # Make the 2nd item have empty string.
  dataset = make_test_data([{'text': 'hello.'}, {'text': ''}, {'text': 'hello world.'}])
  dataset.compute_signal(TestEmbedding(), 'text')

  concept_db = DiskConceptDB()
  concept_db.create(namespace='test_namespace', name='test_concept', type=SignalInputType.TEXT)
  concept_db.edit(
    'test_namespace',
    'test_concept',
    ConceptUpdate(
      insert=[ExampleIn(label=False, text='hello.'), ExampleIn(label=True, text='hello world.')]
    ),
  )

  udf = ConceptSignal(
    namespace='test_namespace', concept_name='test_concept', embedding='test_embedding'
  )
  result = dataset.select_rows(
    columns=['text', Column('text', signal_udf=udf, alias='udf')],
    sort_by=['udf'],
    sort_order=SortOrder.ASC,
  )
  assert list(result) == [
    {'text': '', 'udf': []},
    {
      'text': 'hello.',
      'udf': [{SPAN_KEY: {'start': 0, 'end': 6}, 'score': approx(0.142, abs=1e-3)}],
    },
    {
      'text': 'hello world.',
      'udf': [{SPAN_KEY: {'start': 0, 'end': 12}, 'score': approx(0.958, abs=1e-3)}],
    },
  ]



"""Parquet source."""
import os
import random
from typing import ClassVar, Optional, cast

import duckdb
import pyarrow as pa
from pydantic import Field, ValidationInfo, field_validator
from typing_extensions import override

from ..data import dataset_utils
from ..schema import PARQUET_FILENAME_PREFIX, ROWID, Schema, arrow_schema_to_schema
from ..source import Source, SourceManifest, SourceSchema
from ..sources.duckdb_utils import convert_path_to_duckdb, duckdb_setup
from ..tasks import TaskId
from ..utils import download_http_files

# Number of rows to read per batch.
ROWS_PER_BATCH_READ = 50_000


class ParquetSource(Source):
  """Parquet data loader

  Parquet files can live locally as a filepath, or remotely on GCS, S3, or Hadoop.

  For more details on authentication with private objects, see:
  https://arrow.apache.org/docs/python/filesystems.html
  """  # noqa: D415, D400

  name: ClassVar[str] = 'parquet'
  filepaths: list[str] = Field(
    description='A list of paths to parquet files which live locally or remotely on GCS, S3, or '
    'Hadoop.'
  )
  seed: Optional[int] = Field(description='Random seed for sampling', default=None)
  sample_size: Optional[int] = Field(
    title='Sample size', description='Number of rows to sample from the dataset', default=None
  )
  pseudo_shuffle: bool = Field(
    default=False,
    description='If true, the reader will read a fraction of rows from each shard, '
    'avoiding a pass over the entire dataset.',
  )
  pseudo_shuffle_num_shards: int = Field(
    default=10, description='Number of shards to sample from when using pseudo shuffle.'
  )

  _source_schema: Optional[SourceSchema] = None
  _readers: list[pa.RecordBatchReader] = []
  _con: Optional[duckdb.DuckDBPyConnection] = None

  @field_validator('filepaths')
  @classmethod
  def validate_filepaths(cls, filepaths: list[str]) -> list[str]:
    """Validate filepaths."""
    if not filepaths:
      raise ValueError('filepaths must be non-empty.')
    return filepaths

  @field_validator('sample_size')
  @classmethod
  def validate_sample_size(cls, sample_size: Optional[int]) -> Optional[int]:
    """Validate sample size."""
    if sample_size is not None and sample_size < 1:
      raise ValueError('sample_size must be greater than 0.')
    return sample_size

  @field_validator('pseudo_shuffle')
  @classmethod
  def validate_pseudo_shuffle(cls, pseudo_shuffle: bool, info: ValidationInfo) -> bool:
    """Validate shuffle before sampling."""
    if pseudo_shuffle and not info.data['sample_size']:
      raise ValueError('`pseudo_shuffle` requires `sample_size` to be set.')
    return pseudo_shuffle

  def _setup_sampling(self, duckdb_paths: list[str]) -> None:
    assert self._con, 'setup() must be called first.'
    if self.pseudo_shuffle:
      assert self.sample_size, 'pseudo_shuffle requires sample_size to be set.'
      # Resolve globs into individual files.
      glob_rows: list[tuple[str]] = self._con.execute(
        f'SELECT * FROM GLOB({duckdb_paths})'
      ).fetchall()
      duckdb_files: list[str] = list(set([row[0] for row in glob_rows]))
      # Sub-sample shards so we don't open too many files.
      num_shards = min(self.pseudo_shuffle_num_shards, len(duckdb_files))
      # The 1.5 multiplier gives some wiggle room for heterogeneous shard sizes, in case one shard
      # fails to yield as many samples as hoped.
      samples_per_shard = max(1, int((self.sample_size * 1.5) // num_shards))
      duckdb_files = random.sample(duckdb_files, num_shards)
      self._duckdb_files = duckdb_files

      shard_query = '\nUNION ALL\n'.join(
        f'(SELECT * FROM read_parquet("{path}") LIMIT {samples_per_shard})' for path in duckdb_files
      )
      self._process_query = f"""
          SELECT replace(CAST(uuid() AS VARCHAR), '-', '') AS {ROWID}, *
          FROM ({shard_query}) LIMIT {self.sample_size}"""
    else:
      sample_suffix = ''
      if self.sample_size:
        sample_suffix = f'USING SAMPLE {self.sample_size}'
        if self.seed is not None:
          sample_suffix += f' (reservoir, {self.seed})'
      self._process_query = f"""
          SELECT replace(CAST(uuid() AS VARCHAR), '-', '') AS {ROWID},
              * FROM read_parquet({duckdb_paths}) {sample_suffix}"""

  @override
  def setup(self) -> None:
    filepaths = download_http_files(self.filepaths)
    self._con = duckdb.connect(database=':memory:')
    duckdb_setup(self._con)

    # DuckDB expects s3 protocol: https://duckdb.org/docs/guides/import/s3_import.html.
    duckdb_paths = [convert_path_to_duckdb(path) for path in filepaths]
    schema = arrow_schema_to_schema(
      self._con.execute(f"""SELECT * FROM read_parquet('{duckdb_paths[0]}')""")
      .fetch_record_batch(1)
      .schema
    )
    self._source_schema = SourceSchema(fields=schema.fields, num_items=None)
    self._setup_sampling(duckdb_paths)

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    assert self._source_schema is not None, 'setup() must be called first.'
    return self._source_schema

  @override
  def load_to_parquet(self, output_dir: str, task_id: Optional[TaskId] = None) -> SourceManifest:
    del task_id
    assert self._con, 'setup() must be called first.'

    out_filename = dataset_utils.get_parquet_filename(PARQUET_FILENAME_PREFIX, 0, 1)
    filepath = os.path.join(output_dir, out_filename)
    os.makedirs(output_dir, exist_ok=True)

    self._con.sql(self._process_query).write_parquet(filepath, compression='zstd')

    # Ensure that self.sample_size reflects the actual number of rows processed.
    processed_count = self._con.execute(
      f'SELECT COUNT(*) FROM read_parquet({filepath!r})'
    ).fetchone()
    num_items = cast(tuple[int], processed_count)[0]
    if self.sample_size:
      self.sample_size = min(self.sample_size, num_items)

    schema = Schema(fields=self.source_schema().fields.copy())
    return SourceManifest(files=[out_filename], data_schema=schema, source=self)



"""Tests the spacy chunk splitter."""

from .spacy_splitter import clustering_spacy_chunker, simple_spacy_chunker
from .text_splitter_test_utils import clean_textchunks, text_to_textchunk


def test_short_snippets_filtered() -> None:
  text = '1. Hello. 2. World.'
  expected_spans = text_to_textchunk(text, ['Hello.', 'World.'])

  split_items = simple_spacy_chunker(text)
  assert split_items == expected_spans


def test_colon_considered_as_splitter() -> None:
  text = 'Teacher: Tell me the answer. Student: I have no idea.'
  expected_spans = text_to_textchunk(
    text, ['Teacher:', 'Tell me the answer.', 'Student:', 'I have no idea.']
  )
  split_items = simple_spacy_chunker(text)
  assert split_items == expected_spans


def test_long_spans_default_split() -> None:
  text = 'Blah blah blah.'
  expected_spans = text_to_textchunk(text, ['Blah bl', 'ah blah.'])

  split_items = clustering_spacy_chunker(text, max_len=8)
  assert split_items == expected_spans


def test_long_spans_preferred_splits() -> None:
  text = 'Blah. blah. bla. bl.'
  expected_spans = text_to_textchunk(text, ['Blah.', 'blah.', 'bla.', 'bl.'])
  # Even though target_num_groups = 1, the max len constraint causes breaking.
  split_items = clustering_spacy_chunker(text, target_num_groups=1, max_len=6, filter_short=1)
  assert clean_textchunks(split_items) == expected_spans


def test_similar_spans_grouped() -> None:
  text = 'Blah1. Blah2. Blah2.'
  expected_spans = text_to_textchunk(text, ['Blah1.', 'Blah2. Blah2.'])
  split_items = clustering_spacy_chunker(text, target_num_groups=2)
  assert clean_textchunks(split_items) == expected_spans


def test_fewer_than_desired_groups() -> None:
  text = 'blah blah blah'
  expected_spans = text_to_textchunk(text, ['blah blah blah'])
  split_items = clustering_spacy_chunker(text, target_num_groups=4)
  assert clean_textchunks(split_items) == expected_spans


def test_no_chunks_emitted() -> None:
  text = ''
  split_items = clustering_spacy_chunker(text, target_num_groups=4)
  assert split_items == []



"""Test our public dataset REST API."""
import os
from typing import ClassVar, Iterable, Iterator, Optional, Type

import pytest
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from .auth import UserInfo, get_session_user
from .config import DatasetSettings
from .data.dataset import Dataset, DatasetManifest, SelectRowsSchemaResult, SelectRowsSchemaUDF
from .data.dataset_duckdb import DatasetDuckDB
from .data.dataset_test_utils import (
  TEST_DATASET_NAME,
  TEST_NAMESPACE,
  TestSource,
  enriched_item,
  make_dataset,
)
from .router_dataset import (
  AddLabelsOptions,
  Column,
  SelectRowsOptions,
  SelectRowsResponse,
  SelectRowsSchemaOptions,
  WebManifest,
)
from .schema import Field, Item, RichData, field, schema
from .server import app
from .signal import TextSignal, clear_signal_registry, register_signal
from .source import clear_source_registry, register_source

client = TestClient(app)

DATASET_CLASSES = [DatasetDuckDB]

TEST_DATA: list[Item] = [
  {
    'erased': False,
    'people': [
      {
        'name': 'A',
        'zipcode': 0,
        'locations': [{'city': 'city1', 'state': 'state1'}, {'city': 'city2', 'state': 'state2'}],
      }
    ],
  },
  {
    'erased': True,
    'people': [
      {
        'name': 'B',
        'zipcode': 1,
        'locations': [{'city': 'city3', 'state': 'state3'}, {'city': 'city4'}, {'city': 'city5'}],
      },
      {'name': 'C', 'zipcode': 2, 'locations': [{'city': 'city1', 'state': 'state1'}]},
    ],
  },
  {'erased': True},
]


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_source(TestSource)
  register_signal(LengthSignal)
  # Unit test runs.
  yield
  # Teardown.
  clear_source_registry()
  clear_signal_registry()


@pytest.fixture(scope='module', autouse=True, params=DATASET_CLASSES)
def test_data(
  tmp_path_factory: pytest.TempPathFactory,
  module_mocker: MockerFixture,
  request: pytest.FixtureRequest,
) -> None:
  tmp_path = tmp_path_factory.mktemp('data')
  module_mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})

  dataset_cls: Type[Dataset] = request.param
  make_dataset(dataset_cls, tmp_path, TEST_DATA)


def test_get_manifest() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}'
  response = client.get(url)
  assert response.status_code == 200
  assert WebManifest.model_validate(response.json()) == WebManifest(
    dataset_manifest=DatasetManifest(
      namespace=TEST_NAMESPACE,
      dataset_name=TEST_DATASET_NAME,
      data_schema=schema(
        {
          'erased': 'boolean',
          'people': [
            {
              'name': 'string',
              'zipcode': 'int32',
              'locations': [{'city': 'string', 'state': 'string'}],
            }
          ],
        }
      ),
      num_items=3,
      source=TestSource(),
    )
  )


def test_select_rows_no_options() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/select_rows'
  options = SelectRowsOptions()
  response = client.post(url, json=options.model_dump())
  assert response.status_code == 200
  assert response.json() == {
    'rows': [
      {
        'erased': False,
        'people.*.name': ['A'],
        'people.*.zipcode': [0],
        'people.*.locations.*.city': [['city1', 'city2']],
        'people.*.locations.*.state': [['state1', 'state2']],
      },
      {
        'erased': True,
        'people.*.name': ['B', 'C'],
        'people.*.zipcode': [1, 2],
        'people.*.locations.*.city': [['city3', 'city4', 'city5'], ['city1']],
        'people.*.locations.*.state': [['state3', None, None], ['state1']],
      },
      {'erased': True},
    ],
    'total_num_rows': 3,
  }


def test_select_rows_with_cols_and_limit() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/select_rows'
  options = SelectRowsOptions(
    columns=[('people', '*', 'zipcode'), ('people', '*', 'locations', '*', 'city')],
    limit=1,
    offset=1,
  )
  response = client.post(url, json=options.model_dump())
  assert response.status_code == 200
  assert SelectRowsResponse.model_validate(response.json()) == SelectRowsResponse(
    rows=[
      {
        'people.*.zipcode': [1, 2],
        'people.*.locations.*.city': [['city3', 'city4', 'city5'], ['city1']],
      }
    ],
    total_num_rows=3,
  )


def test_select_rows_with_cols_and_combine() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/select_rows'
  options = SelectRowsOptions(
    columns=[('people', '*', 'zipcode'), ('people', '*', 'locations', '*', 'city')],
    combine_columns=True,
  )
  response = client.post(url, json=options.model_dump())
  assert response.status_code == 200
  assert SelectRowsResponse.model_validate(response.json()) == SelectRowsResponse(
    rows=[
      {'people': [{'zipcode': 0, 'locations': [{'city': 'city1'}, {'city': 'city2'}]}]},
      {
        'people': [
          {'zipcode': 1, 'locations': [{'city': 'city3'}, {'city': 'city4'}, {'city': 'city5'}]},
          {'zipcode': 2, 'locations': [{'city': 'city1'}]},
        ]
      },
      {},
    ],
    total_num_rows=3,
  )


class LengthSignal(TextSignal):
  name: ClassVar[str] = 'length_signal'

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield len(text_content) if text_content is not None else None


def test_select_rows_star_plus_udf() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/select_rows'
  udf = Column(path=('people', '*', 'name'), alias='len', signal_udf=LengthSignal())
  options = SelectRowsOptions(columns=['*', udf], combine_columns=True)
  response = client.post(url, json=options.model_dump())
  assert response.status_code == 200
  assert SelectRowsResponse.model_validate(response.json()) == SelectRowsResponse(
    rows=[
      {
        'erased': False,
        'people': [
          {
            'name': enriched_item('A', {'length_signal': 1}),
            'zipcode': 0,
            'locations': [
              {'city': 'city1', 'state': 'state1'},
              {'city': 'city2', 'state': 'state2'},
            ],
          }
        ],
      },
      {
        'erased': True,
        'people': [
          {
            'name': enriched_item('B', {'length_signal': 1}),
            'zipcode': 1,
            'locations': [
              {'city': 'city3', 'state': 'state3'},
              {'city': 'city4'},
              {'city': 'city5'},
            ],
          },
          {
            'name': enriched_item('C', {'length_signal': 1}),
            'zipcode': 2,
            'locations': [{'city': 'city1', 'state': 'state1'}],
          },
        ],
      },
      {'erased': True},
    ],
    total_num_rows=3,
  )


def test_select_rows_schema_star_plus_udf() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/select_rows_schema'
  signal = LengthSignal()
  udf = Column(path=('people', '*', 'name'), alias='len', signal_udf=signal)
  options = SelectRowsSchemaOptions(columns=['*', udf], combine_columns=True)
  response = client.post(url, json=options.model_dump())
  assert response.status_code == 200
  assert SelectRowsSchemaResult.model_validate(response.json()) == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'erased': 'boolean',
        'people': [
          {
            'name': field(
              'string',
              fields={'length_signal': field('int32', signal.model_dump(exclude_none=True))},
            ),
            'zipcode': 'int32',
            'locations': [{'city': 'string', 'state': 'string'}],
          }
        ],
      }
    ),
    udfs=[SelectRowsSchemaUDF(path=('people', '*', 'name', 'length_signal'), alias='len')],
  )


def test_select_rows_schema_no_cols() -> None:
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/select_rows_schema'
  options = SelectRowsSchemaOptions(combine_columns=True)
  response = client.post(url, json=options.model_dump())
  assert response.status_code == 200
  assert SelectRowsSchemaResult.model_validate(response.json()) == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'erased': 'boolean',
        'people': [
          {
            'name': 'string',
            'zipcode': 'int32',
            'locations': [{'city': 'string', 'state': 'string'}],
          }
        ],
      }
    )
  )


def test_update_settings_auth(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})

  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/settings'
  response = client.post(url, json=DatasetSettings().model_dump())
  assert response.status_code == 401
  assert response.is_error is True
  assert 'User does not have access to update the settings of this dataset.' in response.text


def test_add_labels_auth(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})

  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/labels'
  response = client.post(url, json=AddLabelsOptions(label_name='test_label').model_dump())
  assert response.status_code == 401
  assert response.is_error is True
  assert 'User does not have access to add labels to this dataset.' in response.text

  # Set the user as an admin, they should now be able to create the labels.
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ADMIN_EMAILS': 'admin@test.com,test2@test2.com'})

  # Override the session user so we make them an admin.
  def admin_user() -> UserInfo:
    return UserInfo(
      id='1', email='admin@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = admin_user

  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/labels'
  response = client.post(url, json=AddLabelsOptions(label_name='test_label').model_dump())
  assert response.status_code == 200
  assert response.is_error is False

  # Label again to make sure that existing labels can be written.
  response = client.post(url, json=AddLabelsOptions(label_name='test_label').model_dump())
  assert response.status_code == 200
  assert response.is_error is False


def test_add_labels_auth_user_edit_labels(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ADMIN_EMAILS': 'admin@test.com,test2@test2.com'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_USER_EDIT_LABELS': 'True'})

  # Override the session user so we make them an admin.
  def admin_user() -> UserInfo:
    return UserInfo(
      id='1', email='admin@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = admin_user

  # Create the test_label as an administrator.
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/labels'
  response = client.post(url, json=AddLabelsOptions(label_name='test_label').model_dump())
  assert response.status_code == 200

  # Make the user a non-admin user.
  def user() -> UserInfo:
    return UserInfo(
      id='1', email='user@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = user

  # The user can label results with the same label.
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/labels'
  response = client.post(url, json=AddLabelsOptions(label_name='test_label').model_dump())
  assert response.status_code == 200

  # THe user cannot create new label types.
  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/labels'
  response = client.post(url, json=AddLabelsOptions(label_name='new_label').model_dump())
  assert response.status_code == 401
  assert response.is_error is True
  assert 'User does not have access to create label types in this dataset.' in response.text



"""Router for Google OAuth2 login."""

from urllib.parse import urlparse, urlunparse

from authlib.integrations.starlette_client import OAuth, OAuthError
from fastapi import APIRouter, Request, Response
from fastapi.responses import HTMLResponse
from starlette.config import Config
from starlette.responses import RedirectResponse

from .auth import UserInfo
from .env import env
from .router_utils import RouteErrorHandler

router = APIRouter(route_class=RouteErrorHandler)

if env('LILAC_AUTH_ENABLED'):
  oauth = OAuth(
    Config(
      environ={
        'GOOGLE_CLIENT_ID': env('GOOGLE_CLIENT_ID'),
        'GOOGLE_CLIENT_SECRET': env('GOOGLE_CLIENT_SECRET'),
      }
    )
  )
  oauth.register(
    name='google',
    server_metadata_url='https://accounts.google.com/.well-known/openid-configuration',
    client_kwargs={'scope': 'openid email profile'},
  )


@router.get('/login')
async def login(request: Request, origin_url: str) -> RedirectResponse:
  """Redirects to Google OAuth login page."""
  auth_path = urlunparse(urlparse(origin_url)._replace(path='/google/auth'))
  return await oauth.google.authorize_redirect(request, auth_path)


@router.get('/auth')
async def auth(request: Request) -> Response:
  """Handles the Google OAuth callback."""
  try:
    token = await oauth.google.authorize_access_token(request)
  except OAuthError as error:
    return HTMLResponse(f'<h1>{error}</h1>')
  userinfo = token['userinfo']
  request.session['user'] = UserInfo(
    id=str(userinfo['sub']),
    email=userinfo['email'],
    name=userinfo['name'],
    given_name=userinfo['given_name'],
    family_name=userinfo['family_name'],
  ).model_dump()

  return RedirectResponse(url='/')


@router.get('/logout')
def logout(request: Request) -> RedirectResponse:
  """Logs the user out."""
  request.session.pop('user', None)
  return RedirectResponse(url='/')



"""Utilities for download from a Lilac-processed HuggingFace dataset."""


import os
import shutil
from typing import Optional

from ..config import DatasetConfig, get_dataset_config
from ..db_manager import get_dataset, list_datasets
from ..env import env, get_project_dir
from ..project import add_project_dataset_config, read_project_config
from ..sources.huggingface_source import HuggingFaceSource
from ..utils import (
  get_dataset_output_dir,
  get_datasets_dir,
  get_lilac_cache_dir,
  log,
  read_yaml,
  to_yaml,
)
from .dataset import config_from_dataset
from .dataset_duckdb import DUCKDB_CACHE_FILE


def download(
  url_or_repo: str,
  project_dir: Optional[str] = None,
  dataset_namespace: Optional[str] = 'local',
  dataset_name: Optional[str] = None,
  hf_token: Optional[str] = None,
  overwrite: Optional[bool] = False,
) -> None:
  """Download a Lilac dataset from HuggingFace.

  Args:
    url_or_repo: A remote URL to a Lilac-processed dataset. Currently only supports HuggingFace
      dataset URLs. Can be a full URL: https://huggingface.co/datasets/lilacai/lilac-OpenOrca-100k
      or a repo_id: lilacai/lilac-OpenOrca-100k.
    project_dir: The project directory to use for the demo. Defaults to `env.LILAC_PROJECT_DIR`
      which can be set with `ll.set_project_dir()`.
    dataset_namespace: The local namespace to use. Defaults to 'local'.
    dataset_name: The local dataset name to use. Defaults to the name of the HuggingFace dataset.
    hf_token: The HuggingFace access token to use when making datasets private. This can also be set
      via the `HF_ACCESS_TOKEN` environment flag.
    overwrite: Whether to overwrite the dataset if it already exists.
  """
  try:
    from huggingface_hub import snapshot_download
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  repo_id = url_or_repo.removeprefix('https://huggingface.co/datasets/')

  project_dir = project_dir or get_project_dir()
  datasets_dir = get_datasets_dir(project_dir)
  cache_dir = get_lilac_cache_dir(project_dir)
  os.makedirs(datasets_dir, exist_ok=True)

  print(f'Downloading "{repo_id}" from HuggingFace to {datasets_dir}')
  repo_subdir = os.path.join(cache_dir, 'hf_download', repo_id)
  snapshot_download(
    repo_id=repo_id,
    repo_type='dataset',
    token=env('HF_ACCESS_TOKEN', hf_token),
    local_dir=repo_subdir,
    ignore_patterns=['.gitattributes', 'README.md'],
    local_dir_use_symlinks=False,
  )

  # Find the name of the remote namespace and dataset name.
  remote_namespace = [
    d for d in os.listdir(repo_subdir) if os.path.isdir(os.path.join(repo_subdir, d))
  ][0]
  remote_name = os.listdir(os.path.join(repo_subdir, remote_namespace))[0]
  src = os.path.join(repo_subdir, remote_namespace, remote_name)

  dataset_namespace = dataset_namespace or remote_namespace
  dataset_name = dataset_name or remote_name

  # Check the dataset doesn't already exist.
  datasets = list_datasets(project_dir)
  for dataset in datasets:
    if dataset.namespace == dataset_namespace and dataset.dataset_name == dataset_name:
      if not overwrite:
        raise ValueError(
          f'Dataset {dataset_namespace}/{dataset_name} already exists. '
          'Use --overwrite to overwrite it.'
        )
      else:
        print(f'Overwriting existing dataset {dataset_namespace}/{dataset_name}')
        shutil.rmtree(os.path.join(datasets_dir, dataset_namespace, dataset_name))

  dataset_dir = os.path.join(datasets_dir, dataset_namespace, dataset_name)
  os.makedirs(dataset_dir, exist_ok=True)

  shutil.copytree(src, dataset_dir, dirs_exist_ok=True)

  ds = get_dataset(dataset_namespace, dataset_name)
  dataset_config_filename = os.path.join(dataset_dir, 'dataset_config.yml')
  if os.path.exists(dataset_config_filename):
    dataset_config = DatasetConfig(**read_yaml(dataset_config_filename))
  else:
    dataset_config = config_from_dataset(ds)

  add_project_dataset_config(dataset_config, project_dir, overwrite=True)

  print('Wrote dataset to', dataset_dir)


def upload(
  dataset: str,
  project_dir: Optional[str] = None,
  url_or_repo: Optional[str] = None,
  public: Optional[bool] = False,
  readme_suffix: Optional[str] = None,
  hf_token: Optional[str] = None,
) -> None:
  """Uploads local datasets to HuggingFace datasets.

  Args:
    project_dir: The project directory to use for the demo. Defaults to `env.LILAC_PROJECT_DIR`
      which can be set with `ll.set_project_dir()`.
    dataset: The dataset to upload. Can be a local dataset name, or a namespace/dataset name.
    url_or_repo: The HuggingFace dataset repo ID to use. If not specified, will be automatically
      generated.
    public: Whether to make the dataset public. Defaults to False.
    readme_suffix: A suffix to add to the README.md file. Defaults to None.
    hf_token: The HuggingFace access token to use when making datasets private. This can also be set
      via the `HF_ACCESS_TOKEN` environment flag.
  """
  if not public:
    public = False
  project_dir = project_dir or get_project_dir()

  try:
    from huggingface_hub import HfApi

  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )
  hf_api = HfApi(token=env('HF_ACCESS_TOKEN', hf_token))

  log('Uploading dataset: ', dataset)

  namespace, name = dataset.split('/')
  if url_or_repo:
    repo_id = url_or_repo.removeprefix('https://huggingface.co/datasets/')
  else:
    repo_id = _dataset_repo_id(dataset)

  dataset_link = f'https://huggingface.co/datasets/{repo_id}'
  print(f'Uploading "{dataset}" to HuggingFace dataset repo ' f'{dataset_link}')

  hf_api.create_repo(
    repo_id,
    repo_type='dataset',
    private=not public,
    exist_ok=True,
  )
  dataset_output_dir = get_dataset_output_dir(project_dir, namespace, name)
  hf_api.upload_folder(
    folder_path=dataset_output_dir,
    # The path in the remote doesn't os.path.join as it is specific to Linux.
    path_in_repo=f'{namespace}/{name}',
    repo_id=repo_id,
    repo_type='dataset',
    # Delete all data on the server.
    delete_patterns='*',
    ignore_patterns=[DUCKDB_CACHE_FILE + '*'],
  )

  config = read_project_config(project_dir)
  dataset_config = get_dataset_config(config, namespace, name)
  if dataset_config is None:
    raise ValueError(f'Dataset {namespace}/{name} not found in project config.')

  original_dataset_link = ''
  if isinstance(dataset_config.source, HuggingFaceSource):
    original_dataset_link = f'https://huggingface.co/datasets/{dataset_config.source.dataset_name}'

  dataset_config_yaml = to_yaml(
    dataset_config.model_dump(exclude_defaults=True, exclude_none=True, exclude_unset=True)
  )
  hf_api.upload_file(
    path_or_fileobj=dataset_config_yaml.encode(),
    path_in_repo='dataset_config.yml',
    repo_id=repo_id,
    repo_type='dataset',
  )

  readme = (
    '---\n'
    'tags:\n'
    '- Lilac\n'
    '---\n'
    f'# {dataset}\n'
    'This dataset is a [Lilac](http://lilacml.com) processed dataset. '
    + (
      f'Original dataset: [{original_dataset_link}]({original_dataset_link})\n\n'
      if original_dataset_link != ''
      else ''
    )
    + 'To download the dataset to a local directory:\n\n'
    '```bash\n'
    f'lilac download {repo_id}\n'
    '```\n\n'
    'or from python with:\n\n'
    '```py\n'
    f'll.download("{url_or_repo}")\n'
    '```\n\n' + (readme_suffix or '')
  ).encode()
  hf_api.upload_file(
    path_or_fileobj=readme,
    path_in_repo='README.md',
    repo_id=repo_id,
    repo_type='dataset',
  )

  log(
    f'\nDataset "{dataset}" uploaded to {dataset_link}\n\n'
    'Download the dataset with: \n\n'
    f'$: lilac download {repo_id}\n\n'
    'or from python with: \n\n'
    '```py\n'
    f'll.download("{url_or_repo}")\n'
    '```'
  )


def _dataset_repo_id(dataset: str) -> str:
  """Returns the dataset repo ID from a dataset."""
  namespace, dataset_name = dataset.split('/')
  return f'lilac-{namespace}-{dataset_name}'



"""OpenAI embeddings."""
from typing import ClassVar, Optional

import numpy as np
from tenacity import retry, stop_after_attempt, wait_random_exponential
from typing_extensions import override

from ..env import env
from ..schema import Item
from ..signal import TextEmbeddingSignal
from ..splitters.spacy_splitter import clustering_spacy_chunker
from ..tasks import TaskExecutionType
from .embedding import chunked_compute_embedding

API_NUM_PARALLEL_REQUESTS = 10
API_OPENAI_BATCH_SIZE = 128
API_EMBEDDING_MODEL = 'text-embedding-ada-002'
AZURE_NUM_PARALLEL_REQUESTS = 1
AZURE_OPENAI_BATCH_SIZE = 16


class OpenAIEmbedding(TextEmbeddingSignal):
  """Computes embeddings using OpenAI's embedding API.

  <br>**Important**: This will send data to an external server!

  <br>To use this signal, you must get an OpenAI API key from
  [platform.openai.com](https://platform.openai.com/) and add it to your .env.local.

  <br>For details on pricing, see: https://openai.com/pricing.
  """

  name: ClassVar[str] = 'openai'
  display_name: ClassVar[str] = 'OpenAI Embeddings'
  local_batch_size: ClassVar[int] = API_OPENAI_BATCH_SIZE
  local_parallelism: ClassVar[int] = API_NUM_PARALLEL_REQUESTS
  local_strategy: ClassVar[TaskExecutionType] = 'threads'

  @override
  def setup(self) -> None:
    api_key = env('OPENAI_API_KEY')
    azure_api_key = env('AZURE_OPENAI_KEY')
    azure_api_version = env('AZURE_OPENAI_VERSION')
    azure_api_endpoint = env('AZURE_OPENAI_ENDPOINT')

    if not api_key and not azure_api_key:
      raise ValueError('`OPENAI_API_KEY` or `AZURE_OPENAI_KEY` '
                       'environment variables not set, please set one.')
    if api_key and azure_api_key:
      raise ValueError(
        'Both `OPENAI_API_KEY` and `AZURE_OPENAI_KEY` '
        'environment variables are set, please set only one.')

    try:
      import openai

    except ImportError:
      raise ImportError(
        'Could not import the "openai" python package. '
        'Please install it with `pip install openai`.'
      )

    else:

      if api_key:
        self._client = openai.OpenAI(api_key=api_key)
        self._azure = False

      elif azure_api_key:
        self._client = openai.AzureOpenAI(
          api_key=azure_api_key,
          api_version=azure_api_version,
          azure_endpoint=azure_api_endpoint
        )
        self._azure = True
        OpenAIEmbedding.local_batch_size = AZURE_OPENAI_BATCH_SIZE
        OpenAIEmbedding.local_parallelism = AZURE_NUM_PARALLEL_REQUESTS

  @override
  def compute(self, docs: list[str]) -> list[Optional[Item]]:
    """Compute embeddings for the given documents."""

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(10))
    def embed_fn(texts: list[str]) -> list[np.ndarray]:
      # Replace newlines, which can negatively affect performance.
      # See https://github.com/search?q=repo%3Aopenai%2Fopenai-python+replace+newlines&type=code
      texts = [text.replace('\n', ' ') for text in texts]

      response = self._client.embeddings.create(
        input=texts,
        model=API_EMBEDDING_MODEL if not self._azure else env('AZURE_OPENAI_EMBEDDING_MODEL'),
      )
      return [np.array(embedding.embedding, dtype=np.float32) for embedding in response.data]

    return chunked_compute_embedding(
      embed_fn, docs, self.local_batch_size, chunker=clustering_spacy_chunker
    )



"""Lilac CLI."""

from os.path import abspath
from typing import Literal, Optional, Union

import click

from . import __version__
from .concepts.db_concept import DISK_CONCEPT_DB
from .data.dataset_storage_utils import download, upload
from .deploy import deploy_project
from .env import env, get_project_dir
from .hf_docker_start import hf_docker_start
from .load import load
from .project import dir_is_project, init, project_dir_from_args
from .server import start_server


@click.command()
@click.argument('project_dir', default='')
@click.option(
  '--host',
  help='The host address where the web server will listen to.',
  default='127.0.0.1',
  type=str,
)
@click.option('--port', help='The port number of the web-server', type=int, default=5432)
@click.option(
  '--load', help='Load from the project config upon bootup.', type=bool, is_flag=True, default=False
)
def start(project_dir: str, host: str, port: int, load: bool) -> None:
  """Starts the Lilac web server."""
  project_dir = project_dir_from_args(project_dir)
  if not dir_is_project(project_dir):
    value = str(
      click.prompt(
        f'Lilac will create a project in `{abspath(project_dir)}`. Do you want to continue? (y/n)',
        type=str,
      )
    ).lower()
    if value == 'n':
      exit()

  server = start_server(host=host, port=port, open=True, project_dir=project_dir, load=load)
  # Block the server.
  server.thread.join()


@click.command()
@click.argument('project_dir', default='')
def init_command(project_dir: str) -> None:
  """Initialize a Lilac project in a project directory."""
  project_dir = project_dir_from_args(project_dir)
  if not dir_is_project(project_dir):
    value = str(
      click.prompt(
        f'Lilac will create a project in `{abspath(project_dir)}`. Do you want to continue? (y/n)',
        type=str,
      )
    ).lower()
    if value == 'n':
      exit()

  init(project_dir)


@click.command()
@click.argument('project_dir', default='')
@click.option(
  '--config_path',
  type=str,
  help='[Optional] The path to a json or yml file describing the configuration. '
  'The file contents should be an instance of `lilac.Config` or `lilac.DatasetConfig`. '
  'When not defined, uses `LILAC_PROJECT_DIR`/lilac.yml.',
)
@click.option(
  '--overwrite',
  help='When True, runs all data from scratch, overwriting existing data. When false, only'
  'load new datasets, embeddings, and signals.',
  type=bool,
  is_flag=True,
  default=False,
)
def load_command(project_dir: str, config_path: str, overwrite: bool) -> None:
  """Load from a project configuration."""
  project_dir = project_dir or get_project_dir()
  if not project_dir:
    raise ValueError(
      '--project_dir or the environment variable `LILAC_PROJECT_DIR` must be defined.'
    )

  load(project_dir, config_path, overwrite)


@click.command()
def version() -> None:
  """Prints the version of Lilac."""
  print(__version__)


@click.command()
def hf_docker_start_command() -> None:
  """Prepares the binary by downloading datasets for the HuggingFace docker image."""
  hf_docker_start()


@click.command()
@click.option(
  '--project_dir',
  help='The project directory to use for the demo. Defaults to `env.LILAC_PROJECT_DIR`.',
  type=str,
)
@click.option(
  '--hf_space',
  help='The huggingface space. Should be formatted like `SPACE_ORG/SPACE_NAME`.',
  type=str,
  required=True,
)
@click.option('--dataset', help='The name of a dataset to upload', type=str, multiple=True)
@click.option(
  '--make_datasets_public',
  help='When true, sets the huggingface datasets uploaded to public. Defaults to false.',
  is_flag=True,
  default=False,
)
@click.option(
  '--concept',
  help='The name of a concept to upload. By default all lilac/ concepts are uploaded.',
  type=str,
  multiple=True,
)
@click.option(
  '--skip_data_upload',
  help='When true, only uploads the wheel files without any other changes.',
  is_flag=True,
  default=False,
)
@click.option(
  '--skip_concept_upload',
  help='Skip uploading custom concepts.',
  type=bool,
  is_flag=True,
  default=False,
)
@click.option(
  '--create_space',
  help='When True, creates the HuggingFace space if it doesnt exist. The space will be created '
  'with the storage type defined by --hf_space_storage.',
  is_flag=True,
  default=False,
)
@click.option(
  '--load_on_space',
  help='When True, loads the datasets from your project in the space and does not upload data. '
  'NOTE: This could be expensive if your project config locally has embeddings as they will be '
  'recomputed in HuggingFace.',
  is_flag=True,
  default=False,
)
@click.option(
  '--hf_space_storage',
  help='If defined, sets the HuggingFace space persistent storage type. '
  'NOTE: This only actually sets the space storage type when creating the space. '
  'For more details, see https://huggingface.co/docs/hub/spaces-storage',
  type=click.Choice(['small', 'medium', 'large'], case_sensitive=False),
  default=None,
)
@click.option(
  '--hf_token',
  help='The HuggingFace access token to use when making datasets private. '
  'This can also be set via the `HF_ACCESS_TOKEN` environment flag.',
  type=str,
)
def deploy_project_command(
  project_dir: str,
  hf_space: str,
  dataset: Optional[list[str]],
  make_datasets_public: bool,
  concept: Optional[list[str]],
  skip_data_upload: bool,
  skip_concept_upload: bool,
  create_space: bool,
  load_on_space: bool,
  hf_space_storage: Optional[Union[Literal['small'], Literal['medium'], Literal['large']]],
  hf_token: Optional[str],
) -> None:
  """Deploy a project directory to a HuggingFace Space."""
  # When datasets aren't define, set to None so we upload all datasets.
  if not dataset:
    dataset = None
  # When datasets aren't defined, set to None so we upload all datasets.
  if not concept:
    concept = None

  hf_token = hf_token or env('HF_ACCESS_TOKEN')

  deploy_project(
    project_dir=project_dir,
    hf_space=hf_space,
    datasets=dataset,
    concepts=concept,
    make_datasets_public=make_datasets_public,
    skip_data_upload=skip_data_upload,
    skip_concept_upload=skip_concept_upload,
    create_space=create_space,
    load_on_space=load_on_space,
    hf_space_storage=hf_space_storage,
    hf_token=hf_token,
  )


@click.command()
@click.argument(
  'url_or_repo',
  required=True,
)
@click.option(
  '--project_dir',
  help='The project directory to use for the demo. Defaults to `env.LILAC_PROJECT_DIR`.',
  type=str,
)
@click.option(
  '--dataset_namespace',
  help='The local namespace to use. Defaults to "local".',
  type=str,
  default='local',
)
@click.option(
  '--dataset_name',
  help='The local dataset name to use. Defaults to the name of the HuggingFace dataset.',
  type=str,
)
@click.option(
  '--hf_token',
  help='The HuggingFace access token to use when accessing private datasets. '
  'This can also be set via the `HF_ACCESS_TOKEN` environment flag.',
  type=str,
)
@click.option(
  '--overwrite',
  help='When true, overwrites any existing datasets with the same name.',
  is_flag=True,
  default=False,
)
def download_command(
  url_or_repo: str,
  project_dir: Optional[str],
  dataset_namespace: Optional[str],
  dataset_name: Optional[str],
  hf_token: Optional[str],
  overwrite: Optional[bool] = False,
) -> None:
  """Download a Lilac dataset from HuggingFace."""
  download(url_or_repo, project_dir, dataset_namespace, dataset_name, hf_token, overwrite)


@click.command()
@click.argument(
  'dataset',
  required=True,
)
@click.option(
  '--project_dir',
  help='The project directory to use for the demo. Defaults to `env.LILAC_PROJECT_DIR`.',
  type=str,
)
@click.option(
  '--url_or_repo',
  help='The repo id, or full dataset URL, to use for uploading. For example: lilacai/my-dataset.',
  type=str,
)
@click.option(
  '--public',
  help='When true, makes the dataset public.',
  is_flag=True,
  default=False,
)
@click.option(
  '--readme_suffix',
  help='A suffix string for the readme file.',
  type=str,
)
@click.option(
  '--hf_token',
  help='The HuggingFace access token to use when writing private datasets. '
  'This can also be set via the `HF_ACCESS_TOKEN` environment flag.',
  type=str,
)
def upload_command(
  dataset: str,
  project_dir: Optional[str],
  url_or_repo: Optional[str] = None,
  public: Optional[bool] = False,
  readme_suffix: Optional[str] = None,
  hf_token: Optional[str] = None,
) -> None:
  """Upload a Lilac dataset to HuggingFace."""
  upload(dataset, project_dir, url_or_repo, public, readme_suffix, hf_token)


@click.command()
def concepts() -> None:
  """Lists lilac concepts."""
  print(DISK_CONCEPT_DB.list())


@click.group()
def cli() -> None:
  """Lilac CLI."""
  pass


cli.add_command(version)

cli.add_command(init_command, name='init')
cli.add_command(load_command, name='load')
cli.add_command(start)

cli.add_command(deploy_project_command, name='deploy-project')
cli.add_command(hf_docker_start_command, name='hf-docker-start')
cli.add_command(download_command, name='download')
cli.add_command(upload_command, name='upload')
cli.add_command(concepts)

if __name__ == '__main__':
  cli()



"""Tests for the pandas source."""
import os
import pathlib

# mypy: disable-error-code="attr-defined"
from datasets import ClassLabel, Dataset, DatasetDict, Features, Sequence, Value

from ..schema import schema
from ..source import SourceSchema
from ..test_utils import retrieve_parquet_rows
from .huggingface_source import HF_SPLIT_COLUMN, HuggingFaceSource


def test_hf(tmp_path: pathlib.Path) -> None:
  dataset = Dataset.from_list([{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}])

  dataset_name = os.path.join(tmp_path, 'hf-test-dataset')
  dataset.save_to_disk(dataset_name)

  source = HuggingFaceSource(dataset_name=dataset_name, load_from_disk=True)

  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({HF_SPLIT_COLUMN: 'string', 'x': 'int64', 'y': 'string'}).fields, num_items=2
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [
    {HF_SPLIT_COLUMN: 'default', 'x': 1, 'y': 'ten'},
    {HF_SPLIT_COLUMN: 'default', 'x': 2, 'y': 'twenty'},
  ]


def test_hf_splits(tmp_path: pathlib.Path) -> None:
  dataset = Dataset.from_list(
    [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]
  )
  dataset_splits = DatasetDict({'train': dataset, 'test': dataset})

  dataset_name = os.path.join(tmp_path, 'hf-test-dataset')
  dataset_splits.save_to_disk(dataset_name)

  source = HuggingFaceSource(dataset_name=dataset_name, load_from_disk=True, sample_size=2)

  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({HF_SPLIT_COLUMN: 'string', 'x': 'int64', 'y': 'string'}).fields, num_items=4
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [
    {HF_SPLIT_COLUMN: 'train', 'x': 1, 'y': 'ten'},
    {HF_SPLIT_COLUMN: 'train', 'x': 2, 'y': 'twenty'},
    {HF_SPLIT_COLUMN: 'test', 'x': 1, 'y': 'ten'},
    {HF_SPLIT_COLUMN: 'test', 'x': 2, 'y': 'twenty'},
  ]


def test_hf_sequence(tmp_path: pathlib.Path) -> None:
  dataset = Dataset.from_list(
    [
      {'scalar': 1, 'seq': [1, 0], 'seq_dict': {'x': [1, 2, 3], 'y': ['four', 'five', 'six']}},
      {
        'scalar': 2,
        'seq': [2, 0],
        'seq_dict': {'x': [10, 20, 30], 'y': ['forty', 'fifty', 'sixty']},
      },
    ],
    features=Features(
      {
        'scalar': Value(dtype='int64'),
        'seq': Sequence(feature=Value(dtype='int64')),
        'seq_dict': Sequence(feature={'x': Value(dtype='int64'), 'y': Value(dtype='string')}),
      }
    ),
  )

  dataset_name = os.path.join(tmp_path, 'hf-test-dataset')
  dataset.save_to_disk(dataset_name)

  source = HuggingFaceSource(dataset_name=dataset_name, load_from_disk=True)

  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema(
      {
        HF_SPLIT_COLUMN: 'string',
        'scalar': 'int64',
        'seq': ['int64'],
        'seq_dict': {'x': ['int64'], 'y': ['string']},
      }
    ).fields,
    num_items=2,
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [
    {
      HF_SPLIT_COLUMN: 'default',
      'scalar': 1,
      'seq': [1, 0],
      'seq_dict': {'x': [1, 2, 3], 'y': ['four', 'five', 'six']},
    },
    {
      HF_SPLIT_COLUMN: 'default',
      'scalar': 2,
      'seq': [2, 0],
      'seq_dict': {'x': [10, 20, 30], 'y': ['forty', 'fifty', 'sixty']},
    },
  ]


def test_hf_list(tmp_path: pathlib.Path) -> None:
  dataset = Dataset.from_list(
    [{'scalar': 1, 'list': [{'x': 1, 'y': 'two'}]}, {'scalar': 2, 'list': [{'x': 3, 'y': 'four'}]}],
    features=Features(
      {
        'scalar': Value(dtype='int64'),
        'list': [{'x': Value(dtype='int64'), 'y': Value(dtype='string')}],
      }
    ),
  )

  dataset_name = os.path.join(tmp_path, 'hf-test-dataset')
  dataset.save_to_disk(dataset_name)

  source = HuggingFaceSource(dataset_name=dataset_name, load_from_disk=True)

  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema(
      {HF_SPLIT_COLUMN: 'string', 'scalar': 'int64', 'list': [{'x': 'int64', 'y': 'string'}]}
    ).fields,
    num_items=2,
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [
    {HF_SPLIT_COLUMN: 'default', 'scalar': 1, 'list': [{'x': 1, 'y': 'two'}]},
    {HF_SPLIT_COLUMN: 'default', 'scalar': 2, 'list': [{'x': 3, 'y': 'four'}]},
  ]


def test_hf_class_labels(tmp_path: pathlib.Path) -> None:
  dataset = Dataset.from_list(
    [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}],
    features=Features(
      {
        'x': ClassLabel(num_classes=3, names=['x0', 'x1', 'x2']),
        'y': Value(dtype='string'),
      }
    ),
  )
  dataset_name = os.path.join(tmp_path, 'hf-test-dataset')
  dataset.save_to_disk(dataset_name)

  source = HuggingFaceSource(dataset_name=dataset_name, load_from_disk=True)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({HF_SPLIT_COLUMN: 'string', 'x': 'string', 'y': 'string'}).fields, num_items=2
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [
    {HF_SPLIT_COLUMN: 'default', 'x': 'x1', 'y': 'ten'},
    {HF_SPLIT_COLUMN: 'default', 'x': 'x2', 'y': 'twenty'},
  ]



"""Sources for ingesting data into Lilac."""

from .csv_source import CSVSource
from .default_sources import register_default_sources
from .dict_source import DictSource
from .github_source import GithubSource
from .gmail_source import GmailSource
from .huggingface_source import HuggingFaceSource
from .json_source import JSONSource
from .langsmith import LangSmithSource
from .llama_index_docs_source import LlamaIndexDocsSource
from .pandas_source import PandasSource
from .parquet_source import ParquetSource

register_default_sources()

__all__ = [
  'HuggingFaceSource',
  'CSVSource',
  'JSONSource',
  'GmailSource',
  'PandasSource',
  'ParquetSource',
  'DictSource',
  'LangSmithSource',
  'GithubSource',
  'LlamaIndexDocsSource',
]



"""LangSmith source."""
from typing import ClassVar, Iterable, Optional

from fastapi import APIRouter
from pydantic import Field
from typing_extensions import override

from ..env import env
from ..schema import Item
from ..source import Source, SourceSchema
from .dict_source import DictSource

router = APIRouter()

DEFAULT_LANGCHAIN_ENDPOINT = 'https://api.smith.langchain.com'


@router.get('/datasets')
def get_langsmith_datasets() -> list[str]:
  """List the datasets in LangSmith."""
  from langsmith import Client

  client = Client()
  return [d.name for d in client.list_datasets()]


class LangSmithSource(Source):
  """LangSmith data loader."""

  name: ClassVar[str] = 'langsmith'
  router: ClassVar[APIRouter] = router

  dataset_name: str = Field(description='LangSmith dataset name')

  _dict_source: Optional[DictSource] = None

  @override
  def setup(self) -> None:
    api_key = env('LANGCHAIN_API_KEY')
    api_url = env('LANGCHAIN_ENDPOINT', DEFAULT_LANGCHAIN_ENDPOINT)
    if not api_key or not api_url:
      raise ValueError(
        '`LANGCHAIN_API_KEY` and `LANGCHAIN_ENDPOINT` environment variables must be set.'
      )
    try:
      from langsmith import Client
    except ImportError:
      raise ImportError(
        'Could not import dependencies for the LangSmith source. '
        'Please install the dependency via `pip install lilac[langsmith]`.'
      )
    client = Client(api_key=api_key, api_url=api_url)

    items = [
      {**example.inputs, **(example.outputs or {})}
      for example in client.list_examples(dataset_name=self.dataset_name)
    ]

    self._dict_source = DictSource(items)
    self._dict_source.setup()

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    assert self._dict_source is not None
    return self._dict_source.source_schema()

  @override
  def yield_items(self) -> Iterable[Item]:
    """Process the source."""
    assert self._dict_source is not None
    return self._dict_source.yield_items()



"""Tests for dataset.select_rows(searches=[...])."""

from typing import ClassVar, Iterable, Iterator, cast

import numpy as np
import pytest
from pytest import approx
from pytest_mock import MockerFixture
from typing_extensions import override

from ..concepts.concept import ExampleIn, LogisticEmbeddingModel
from ..concepts.db_concept import ConceptUpdate, DiskConceptDB
from ..db_manager import set_default_dataset_cls
from ..schema import ROWID, Item, RichData, SignalInputType, lilac_embedding, span
from ..signal import TextEmbeddingSignal, clear_signal_registry, register_signal
from ..signals.concept_scorer import ConceptSignal
from ..signals.semantic_similarity import SemanticSimilaritySignal
from ..signals.substring_search import SubstringSignal
from .dataset import ConceptSearch, Filter, KeywordSearch, SemanticSearch, SortOrder
from .dataset_duckdb import DatasetDuckDB
from .dataset_test_utils import TestDataMaker, enriched_item

TEST_DATA: list[Item] = [
  {'text': 'hello world', 'text2': 'again hello world'},
  {'text': 'looking for world in text', 'text2': 'again looking for world in text'},
  {'text': 'unrelated text', 'text2': 'again unrelated text'},
]

EMBEDDINGS: list[tuple[str, list[float]]] = [
  ('hello.', [1.0, 0.0, 0.0]),
  ('hello2.', [1.0, 1.0, 0.0]),
  ('hello world.', [1.0, 1.0, 1.0]),
  ('hello world2.', [2.0, 1.0, 1.0]),
  ('random negative 1', [0, 0, 0.3]),
  ('random negative 2', [0, 0, 0.4]),
  ('random negative 3', [0, 0.1, 0.5]),
  ('random negative 4', [0.1, 0, 0.4]),
]

STR_EMBEDDINGS: dict[str, list[float]] = {text: embedding for text, embedding in EMBEDDINGS}


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  set_default_dataset_cls(DatasetDuckDB)
  register_signal(TestEmbedding)

  # Unit test runs.
  yield

  # Teardown.
  clear_signal_registry()


def test_search_keyword(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  query = 'world'
  result = dataset.select_rows(
    searches=[KeywordSearch(path='text', query=query)], combine_columns=True
  )

  expected_signal_udf = SubstringSignal(query=query)
  assert list(result) == [
    {
      'text': enriched_item('hello world', {expected_signal_udf.key(): [span(6, 11)]}),
      'text2': 'again hello world',
    },
    {
      'text': enriched_item(
        'looking for world in text', {expected_signal_udf.key(): [span(12, 17)]}
      ),
      'text2': 'again looking for world in text',
    },
  ]


def test_search_keyword_special_chars(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': 'This is 100%'}, {'text': 'This has _underscore_'}, {'text': "Let's do this"}]
  )

  query = '100%'
  result = dataset.select_rows(
    searches=[KeywordSearch(path='text', query=query)], combine_columns=True
  )

  expected_signal_udf = SubstringSignal(query=query)
  assert list(result) == [
    {'text': enriched_item('This is 100%', {expected_signal_udf.key(): [span(8, 12)]})}
  ]

  query = '_underscore_'
  result = dataset.select_rows(
    searches=[KeywordSearch(path='text', query=query)], combine_columns=True
  )

  expected_signal_udf = SubstringSignal(query=query)
  assert list(result) == [
    {'text': enriched_item('This has _underscore_', {expected_signal_udf.key(): [span(9, 21)]})}
  ]

  query = "let's"
  result = dataset.select_rows(
    searches=[KeywordSearch(path='text', query=query)], combine_columns=True
  )
  expected_signal_udf = SubstringSignal(query=query)
  assert list(result) == [
    {'text': enriched_item("Let's do this", {expected_signal_udf.key(): [span(0, 5)]})}
  ]


def test_search_keyword_multiple(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  query_world = 'world'
  query_looking_world = 'looking for world'
  expected_world_udf = SubstringSignal(query=query_world)
  expected_again_looking_udf = SubstringSignal(query=query_looking_world)

  result = dataset.select_rows(
    searches=[
      KeywordSearch(path='text', query=query_world),
      KeywordSearch(path='text2', query=query_looking_world),
    ],
    combine_columns=True,
  )

  assert list(result) == [
    {
      'text': enriched_item(
        'looking for world in text', {expected_world_udf.key(): [span(12, 17)]}
      ),
      'text2': enriched_item(
        'again looking for world in text', {expected_again_looking_udf.key(): [span(6, 23)]}
      ),
    }
  ]


def test_search_keyword_with_filters(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  query = 'world'
  result = dataset.select_rows(
    ['*'],
    filters=[(ROWID, 'in', ['1', '3'])],
    searches=[KeywordSearch(path='text', query=query)],
    combine_columns=True,
  )

  expected_signal_udf = SubstringSignal(query=query)
  assert list(result) == [
    {
      'text': enriched_item('hello world', {expected_signal_udf.key(): [span(6, 11)]}),
      'text2': 'again hello world',
    }
    # The second row doesn't match the rowid filter.
  ]


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      embedding = np.array(STR_EMBEDDINGS[cast(str, example)])
      yield [lilac_embedding(0, len(example), embedding)]


def test_semantic_search(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}, {'text': 'hello world2.'}])

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  query = 'hello2.'
  result = dataset.select_rows(
    searches=[SemanticSearch(path='text', query=query, embedding='test_embedding')],
    combine_columns=True,
  )
  expected_signal_udf = SemanticSimilaritySignal(query=query, embedding='test_embedding')
  assert list(result) == [
    # Results are sorted by score desc.
    {
      'text': enriched_item(
        'hello world2.', {expected_signal_udf.key(): [span(0, 13, {'score': 3})]}
      )
    },
    {
      'text': enriched_item(
        'hello world.', {expected_signal_udf.key(): [span(0, 12, {'score': 2})]}
      )
    },
  ]


def test_concept_search(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  concept_model_mock = mocker.spy(LogisticEmbeddingModel, 'fit')

  dataset = make_test_data(
    [
      {'text': 'hello world.'},
      {'text': 'hello world2.'},
      {'text': 'random negative 1'},
      {'text': 'random negative 2'},
      {'text': 'random negative 3'},
      {'text': 'random negative 4'},
    ]
  )

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  concept_db = DiskConceptDB()
  concept_db.create(namespace='test_namespace', name='test_concept', type=SignalInputType.TEXT)
  concept_db.edit(
    'test_namespace',
    'test_concept',
    ConceptUpdate(
      insert=[
        ExampleIn(label=False, text='hello world.'),
        ExampleIn(label=True, text='hello world2.'),
      ]
    ),
  )

  result = dataset.select_rows(
    columns=[ROWID, '*'],
    searches=[
      ConceptSearch(
        path='text',
        concept_namespace='test_namespace',
        concept_name='test_concept',
        embedding='test_embedding',
      )
    ],
    filters=[(ROWID, 'in', ['1', '2'])],
    combine_columns=True,
  )
  expected_signal_udf = ConceptSignal(
    namespace='test_namespace', concept_name='test_concept', embedding='test_embedding'
  )

  assert list(result) == [
    # Results are sorted by score desc.
    {
      ROWID: '2',
      'text': enriched_item(
        'hello world2.',
        {
          expected_signal_udf.key(): [span(0, 13, {'score': approx(0.75, abs=0.25)})],
          'test_namespace/test_concept/labels/preview': [span(0, 13, {'label': True})],
        },
      ),
    },
    {
      ROWID: '1',
      'text': enriched_item(
        'hello world.',
        {
          expected_signal_udf.key(): [span(0, 12, {'score': approx(0.25, abs=0.25)})],
          'test_namespace/test_concept/labels/preview': [span(0, 12, {'label': False})],
        },
      ),
    },
  ]

  (_, embeddings, labels) = concept_model_mock.call_args_list[-1].args
  assert embeddings.shape == (2, 3)
  assert labels == [
    # Explicit labels.
    False,
    True,
  ]


def test_concept_search_without_rowid(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}, {'text': 'hello world2.'}])

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  concept_db = DiskConceptDB()
  concept_db.create(namespace='test_namespace', name='test_concept', type=SignalInputType.TEXT)
  concept_db.edit(
    'test_namespace',
    'test_concept',
    ConceptUpdate(
      insert=[
        ExampleIn(label=False, text='hello world.'),
        ExampleIn(label=True, text='hello world2.'),
      ]
    ),
  )

  result = dataset.select_rows(
    columns=['text'],
    searches=[
      ConceptSearch(
        path='text',
        concept_namespace='test_namespace',
        concept_name='test_concept',
        embedding='test_embedding',
      )
    ],
  )

  assert list(result) == [
    # Results are sorted by score desc.
    {
      'text': 'hello world2.',
      'text.test_namespace/test_concept/test_embedding/preview': [
        span(0, 13, {'score': approx(0.75, abs=0.25)})
      ],
      'text.test_namespace/test_concept/labels/preview': [span(0, 13, {'label': True})],
    },
    {
      'text': 'hello world.',
      'text.test_namespace/test_concept/test_embedding/preview': [
        span(0, 12, {'score': approx(0.25, abs=0.25)})
      ],
      'text.test_namespace/test_concept/labels/preview': [span(0, 12, {'label': False})],
    },
  ]

  result = dataset.select_rows(
    columns=['text'],
    searches=[
      ConceptSearch(
        path='text',
        concept_namespace='test_namespace',
        concept_name='test_concept',
        embedding='test_embedding',
      )
    ],
    combine_columns=True,
  )

  assert list(result) == [
    # Results are sorted by score desc.
    {
      'text': enriched_item(
        'hello world2.',
        {
          'test_namespace/test_concept/test_embedding/preview': [
            span(0, 13, {'score': approx(0.75, abs=0.25)})
          ],
          'test_namespace/test_concept/labels/preview': [span(0, 13, {'label': True})],
        },
      )
    },
    {
      'text': enriched_item(
        'hello world.',
        {
          'test_namespace/test_concept/test_embedding/preview': [
            span(0, 12, {'score': approx(0.25, abs=0.25)})
          ],
          'test_namespace/test_concept/labels/preview': [span(0, 12, {'label': False})],
        },
      )
    },
  ]


def test_concept_search_sort_by_rowid(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}, {'text': 'hello world2.'}])

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  concept_db = DiskConceptDB()
  concept_db.create(namespace='test_namespace', name='test_concept', type=SignalInputType.TEXT)
  concept_db.edit(
    'test_namespace',
    'test_concept',
    ConceptUpdate(
      insert=[
        ExampleIn(label=False, text='hello world.'),
        ExampleIn(label=True, text='hello world2.'),
      ]
    ),
  )

  result = dataset.select_rows(
    columns=['text'],
    searches=[
      ConceptSearch(
        path='text',
        concept_namespace='test_namespace',
        concept_name='test_concept',
        embedding='test_embedding',
      )
    ],
    sort_by=[ROWID],
    sort_order=SortOrder.ASC,
  )

  assert list(result) == [
    # Results are sorted by rowid.
    {
      'text': 'hello world.',
      'text.test_namespace/test_concept/test_embedding/preview': [
        span(0, 12, {'score': approx(0.25, abs=0.25)})
      ],
      'text.test_namespace/test_concept/labels/preview': [span(0, 12, {'label': False})],
    },
    {
      'text': 'hello world2.',
      'text.test_namespace/test_concept/test_embedding/preview': [
        span(0, 13, {'score': approx(0.75, abs=0.25)})
      ],
      'text.test_namespace/test_concept/labels/preview': [span(0, 13, {'label': True})],
    },
  ]


def test_concept_search_over_repeated_string(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': ['hello.', 'hello world.']}, {'text': ['hello2.', 'hello world2.']}]
  )
  dataset.compute_embedding('test_embedding', ('text', '*'))

  concept_db = DiskConceptDB()
  concept_db.create(namespace='test_namespace', name='test_concept', type=SignalInputType.TEXT)
  concept_db.edit(
    'test_namespace',
    'test_concept',
    ConceptUpdate(
      insert=[
        ExampleIn(label=False, text='hello world.'),
        ExampleIn(label=True, text='hello world2.'),
      ]
    ),
  )

  result = dataset.select_rows(
    columns=['text'],
    searches=[
      ConceptSearch(
        path=('text', '*'),
        concept_namespace='test_namespace',
        concept_name='test_concept',
        embedding='test_embedding',
      )
    ],
    filters=[Filter(path=(ROWID,), op='in', value=['1', '2'])],
  )
  assert list(result) == [
    {
      'text': ['hello.', 'hello world.'],
      'text.test_namespace/test_concept/labels/preview': [
        None,
        [{'__span__': {'end': 12, 'start': 0}, 'label': False}],
      ],
      'text.test_namespace/test_concept/test_embedding/preview': [
        [{'__span__': {'end': 6, 'start': 0}, 'score': approx(0.93, abs=0.01)}],
        [{'__span__': {'end': 12, 'start': 0}, 'score': approx(0.26, abs=0.01)}],
      ],
    },
    {
      'text': ['hello2.', 'hello world2.'],
      'text.test_namespace/test_concept/labels/preview': [
        None,
        [{'__span__': {'end': 13, 'start': 0}, 'label': True}],
      ],
      'text.test_namespace/test_concept/test_embedding/preview': [
        [{'__span__': {'end': 7, 'start': 0}, 'score': approx(0.68, abs=0.01)}],
        [{'__span__': {'end': 13, 'start': 0}, 'score': approx(0.82, abs=0.01)}],
      ],
    },
  ]


def test_sort_override_search(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': 'hello world.', 'value': 10}, {'text': 'hello world2.', 'value': 20}]
  )

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  query = 'hello2.'
  search = SemanticSearch(path='text', query=query, embedding='test_embedding')

  expected_signal_udf = SemanticSimilaritySignal(query=query, embedding='test_embedding')
  expected_item_1 = {
    'text': enriched_item(
      'hello world.', {expected_signal_udf.key(): [span(0, 12, {'score': 2.0})]}
    ),
    'value': 10,
  }
  expected_item_2 = {
    'text': enriched_item(
      'hello world2.', {expected_signal_udf.key(): [span(0, 13, {'score': 3.0})]}
    ),
    'value': 20,
  }

  sort_order = SortOrder.ASC
  result = dataset.select_rows(
    searches=[search], sort_by=[('value',)], sort_order=sort_order, combine_columns=True
  )
  assert list(result) == [
    # Results are sorted by score ascending.
    expected_item_1,
    expected_item_2,
  ]

  sort_order = SortOrder.DESC
  result = dataset.select_rows(
    searches=[search], sort_by=[('text',)], sort_order=sort_order, combine_columns=True
  )
  assert list(result) == [
    # Results are sorted by score descending.
    expected_item_2,
    expected_item_1,
  ]


def test_search_keyword_and_semantic(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}, {'text': 'hello world2.'}])

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  query = 'hello2.'
  keyword_query = 'rld2'
  result = dataset.select_rows(
    searches=[
      SemanticSearch(path='text', query=query, embedding='test_embedding'),
      KeywordSearch(path='text', query=keyword_query),
    ],
    combine_columns=True,
  )
  expected_semantic_signal = SemanticSimilaritySignal(query=query, embedding='test_embedding')
  expected_keyword_signal = SubstringSignal(query=keyword_query)
  assert list(result) == [
    # Results are sorted by score desc.
    {
      'text': enriched_item(
        'hello world2.',
        {
          expected_semantic_signal.key(): [span(0, 13, {'score': 3})],
          expected_keyword_signal.key(): [span(8, 12)],
        },
      )
    }
    # rowid '1' is not returned because it does not match the keyword query.
  ]



"""Loads reddit data from Huggingface."""
from typing import ClassVar, Iterable, Optional

from pydantic import Field as PydanticField
from typing_extensions import override

from ..schema import Item
from ..source import Source, SourceSchema
from .huggingface_source import HuggingFaceSource

HF_REDDIT_DATASET_NAME = 'reddit'
HF_SUBREDDIT_COL = 'subreddit'


class RedditDataset(Source):
  """Reddit data loader, using Huggingface.

  Loads data from [huggingface.co/datasets/reddit](https://huggingface.co/datasets/reddit).
  """  # noqa: D415, D400

  name: ClassVar[str] = 'reddit'

  subreddits: Optional[list[str]] = PydanticField(
    description='If defined, only loads the subset of reddit data in these subreddit.'
  )

  _hf_dataset: HuggingFaceSource

  @override
  def setup(self) -> None:
    self._hf_dataset = HuggingFaceSource(dataset_name=HF_REDDIT_DATASET_NAME)
    self._hf_dataset.setup()

  @override
  def source_schema(self) -> SourceSchema:
    return self._hf_dataset.source_schema()

  @override
  def yield_items(self) -> Iterable[Item]:
    items = self._hf_dataset.yield_items()

    if not self.subreddits:
      return items

    lower_subreddits = [subreddit.lower() for subreddit in self.subreddits]

    for item in items:
      item_subreddit = item[HF_SUBREDDIT_COL]
      if item_subreddit.lower() not in lower_subreddits:
        # Yield None so that the progress bar is accurate.
        yield None
        continue
      yield item
